/////////////////////////////////////////////////////////////////////////////
// Name:        hexedit.cpp
// Purpose:     hexadecimal editor control
// Package:     ASIX FORTE programmer demo software
//              distributed as-is, without any warranties
// Author:      http://www.asix.net/
// Created:     2016
/////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

// for all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWidgets headers)
#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

// standard header file
#include "hexedit.h"

// ============================================================================
// implementation
// ============================================================================
 
// ----------------------------------------------------------------------------
// event tables and other macros for wxWidgets
// ----------------------------------------------------------------------------

// the event tables connect the wxWidgets events with the functions (event
// handlers) which process them. It can be also done at run-time, but for the
// simple menu events like this the static method is much simpler.

BEGIN_EVENT_TABLE(HexEdit, wxScrolledWindow)
	EVT_CHAR( HexEdit::OnChar )
	EVT_SIZE( HexEdit::OnSize )
	EVT_PAINT( HexEdit::OnPaint )
	EVT_LEFT_DOWN( HexEdit::OnMouseLeft )
	EVT_MOTION( HexEdit::OnMouseMove )
	EVT_SET_FOCUS( HexEdit::OnFocus )
    EVT_KILL_FOCUS( HexEdit::OnFocus )
    EVT_SCROLLWIN( HexEdit::OnScroll )
END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// class HexEdit
// ----------------------------------------------------------------------------
// constructor

HexEdit::HexEdit(wxWindow *parent, wxWindowID id, int memory_word_count, int bytes_per_mem_word,
        const wxPoint &pos, const wxSize &size, long style)
        : wxScrolledWindow(parent, id, pos, size, style)
{
    int i;

    word_count = memory_word_count;
    bytes_per_word = bytes_per_mem_word;
    bytes_count = word_count*bytes_per_word;
    memory = new unsigned char[bytes_count];
    for(i=0;i<bytes_count;i++) memory[i]=0;
    max_words_per_line = 16;
    wxString last_address;
    last_address.Printf("%x", memory_word_count-1);
    address_digits = last_address.Len();
    address_offset = 0;
    cursor_pos = 0;
    sel_start = 0;
    sel_stop = 0;
    selected = false;
    cursor_in_hex = true;
    update_callback_set = false;
	internalBufferDC = NULL;
	internalBufferBMP = NULL;

    // default font - try to find fixed width font
    wxArrayString prefferedfonts, availablefonts;
    wxString fontfacename;
    prefferedfonts.Add("Courier New");
    prefferedfonts.Add("Courier");
    availablefonts = wxFontEnumerator::GetFacenames(wxFONTENCODING_ISO8859_1, true);
    for(i=0;i<(int)prefferedfonts.Count();i++) {
        if(availablefonts.Index(prefferedfonts[i]) != wxNOT_FOUND) {
            fontfacename = prefferedfonts[i];
            break;
        }
    }
    if(fontfacename.IsEmpty() && (availablefonts.Count()>0)) {
        fontfacename = availablefonts[0];
    }
    font = wxFont(
				10,
				wxFONTFAMILY_MODERN,	// family
				wxFONTSTYLE_NORMAL,	    // style
				wxFONTWEIGHT_NORMAL,      // weight
				false,				    // underline
				fontfacename,			    // facename
				wxFONTENCODING_CP437);// msdos encoding
        
    this->ShowScrollbars(wxSHOW_SB_NEVER, wxSHOW_SB_ALWAYS);

    Resize();

    wxCaret *caret = new wxCaret(this, character_width, character_height);
    SetCaret(caret);
    MoveCaret();
}

HexEdit::~HexEdit()
{
    delete memory;
}

#ifdef HEX_EDIT_H_DEBUG
void HexEdit::DebugMessage(wxString msg)
{
    wxFrame *frame = wxDynamicCast(GetParent(), wxFrame);
    if ( frame && frame->GetStatusBar() ){
        frame->SetStatusText(msg);
    }
}

void HexEdit::DebugMessage(int msg)
{
    wxString s;
    s.Printf("%d (0x%x)", msg, msg);
    DebugMessage(s);
}

void HexEdit::DebugMessage(int msg1, int msg2)
{
    wxString s;
    s.Printf("%d %d", msg1, msg2);
    DebugMessage(s);
}
#endif // HEX_EDIT_H_DEBUG

void HexEdit::OnSize(wxSizeEvent &event)
{
    Resize();
}

void HexEdit::Resize()
{
    wxClientDC dc(this);
    dc.SetFont(font);
    character_width = dc.GetCharWidth();
    character_height = dc.GetCharHeight();
        
    // 00: 3031 3233 3435 | 012345
    wxSize size = GetClientSize();
    int chars_per_line = size.x / character_width;
    words_per_line = (chars_per_line - address_digits - 1 - 3) / (bytes_per_word*3+1);
    if(words_per_line > max_words_per_line) words_per_line = max_words_per_line;
    if(words_per_line < 1) words_per_line = 1;
    bytes_per_line = words_per_line*bytes_per_word;
    address_offset = (address_offset / words_per_line) * words_per_line; // round

    visible_lines = size.y / character_height;
    if(visible_lines < 1) visible_lines = 1;

    int total_lines = word_count / words_per_line;
    if((word_count % words_per_line) != 0) total_lines++;
    total_lines++; // header
    SetScrollbars(1, character_height, 1, total_lines, 0, address_offset/words_per_line, true);
    address_offset_max = (total_lines - visible_lines)*words_per_line;
    while(address_offset > address_offset_max) address_offset -= words_per_line;
    if(address_offset < 0) address_offset = 0;

	CreateDC();
	Repaint();
}

void HexEdit::OnScroll(wxScrollWinEvent &event)
{
    if (event.GetEventType() == wxEVT_SCROLLWIN_TOP) {
        //DebugMessage("top");
        return;
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_BOTTOM) {
        //DebugMessage("bot");
        return;
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_LINEUP) {
        address_offset -= words_per_line;
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_LINEDOWN) {
        address_offset += words_per_line;
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_PAGEUP) {
        address_offset -= words_per_line * (visible_lines-1);
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_PAGEDOWN) {
        address_offset += words_per_line * (visible_lines-1);
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_THUMBTRACK) {
        address_offset = event.GetPosition() * words_per_line;
    }
    else if (event.GetEventType() == wxEVT_SCROLLWIN_THUMBRELEASE) {
        //DebugMessage("th-rel");
        return;
    }
    if(address_offset < 0) address_offset = 0;
    if(address_offset > address_offset_max) address_offset = address_offset_max;
    Scroll(0, address_offset/words_per_line);
    Repaint();
}

void HexEdit::Repaint( void )
{
    MoveCaret();
	wxDC* dcTemp = UpdateDC();
	if(dcTemp != NULL){
		wxClientDC dc(this);
		dc.Blit(0, 0, this->GetSize().GetWidth(), this->GetSize().GetHeight(), dcTemp, 0, 0, wxCOPY);
	}
}

void HexEdit::OnPaint(wxPaintEvent &WXUNUSED(event))
{
    wxDC* dcTemp = UpdateDC();
	if(dcTemp != NULL){
		wxPaintDC dc(this);
		dc.Blit(0, 0, this->GetSize().GetWidth(), this->GetSize().GetHeight(), dcTemp, 0, 0, wxCOPY);
	}
}

wxMemoryDC* HexEdit::CreateDC()
{
	if(internalBufferDC != NULL)
		delete internalBufferDC;
	if(internalBufferBMP != NULL)
		delete internalBufferBMP;
	internalBufferBMP=new wxBitmap(this->GetSize().GetWidth(), this->GetSize().GetHeight());
	internalBufferDC = new wxMemoryDC();
	internalBufferDC->SelectObject(*internalBufferBMP);
	return internalBufferDC;
}

wxDC* HexEdit::UpdateDC()
{
	if(internalBufferDC==NULL)
		internalBufferDC = CreateDC();
	wxMemoryDC *dcTemp = internalBufferDC;
    dcTemp->SetFont(font);

    // clear canvas
	wxBrush dbrush(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW));
	dcTemp->SetBackground(dbrush);
	dcTemp->SetBackgroundMode(wxSOLID);
	dcTemp->Clear();

    // word x position
    int i,j, aoff, y;
    int justlen;
    wxString s, buf, just;
    s = " ";
    for(i=0;i<address_digits;i++) s+=" ";

    justlen = bytes_per_word*2-2;
    just = wxString();
    for(i=0;i<justlen;i++) just+=" ";
    for(i=0;i<words_per_line;i++) {
        s += " ";
        buf.Printf("%02x",i & 0xFF);
        s += buf + just;
    }
    s += " | ";
    if(bytes_per_word<=2) {
        justlen = bytes_per_word-1;
        just = wxString();
        for(i=0;i<justlen;i++) just+=" ";
        for(i=0;i<words_per_line;i++) {
            buf.Printf("%x",i & 0xF);
            s += buf + just;
        }
    }
    else {
        justlen = bytes_per_word-2;
        just = wxString();
        for(i=0;i<justlen;i++) just+=" ";
        for(i=0;i<words_per_line;i++) {
            buf.Printf("%02x",i & 0xFF);
            s += buf + just;
        }
    }
    dcTemp->SetTextForeground(wxColour(0xFF, 0xFF, 0xFF));
    dcTemp->SetTextBackground(wxColour(0x80, 0x80, 0x80));
    dcTemp->DrawText(s, 0, 0);

    // word y position
    aoff = address_offset;
    y = 1;
    while(aoff<word_count) {
        s.Printf("%%0%dx:", address_digits);
        buf.Printf(s, aoff);
        dcTemp->DrawText(buf, 0, character_height*y);
        y++;
        aoff += words_per_line;
        if(y>=visible_lines) break;
    }

    // zebra
    s = "";
    for(i=0;i<words_per_line;i++) {
        s += " ";
        for(j=0;j<bytes_per_word;j++) {
            s += "  ";
        }
    }
    int hexlen = s.Len();
    s += " | ";
    for(i=0;i<bytes_per_line;i++) {
        s += " ";
    }
    
    dcTemp->SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT));
    int lines_above;
    aoff = address_offset;
    y = 1;
    while(aoff<word_count) {
        lines_above = aoff / words_per_line;
        if((lines_above % 2) == 0) dcTemp->SetTextBackground(wxColour(0xC9, 0xEA, 0xFD));
        else dcTemp->SetTextBackground(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW));
        dcTemp->DrawText(s, character_width*(address_digits+1), character_height*y);
        y++;
        aoff += words_per_line;
        if(y>=visible_lines) break;
    }
    
    // data
    int sel_start_sort = sel_start;
    int sel_stop_sort = sel_stop;
    if(sel_start > sel_stop) {
        sel_start_sort = sel_stop;
        sel_stop_sort = sel_start;
    }
    sel_start_sort /= 2; // nibbles->bytes
    sel_stop_sort /= 2;
    aoff = address_offset;
    y = 1;
    while(aoff<word_count) {
        bool whole_line_selected = false;
        bool whole_line_not_selected = false;
        int line_first_byte = aoff * bytes_per_word;
        int line_last_byte = line_first_byte + bytes_per_line - 1;
        if(selected && (sel_start_sort<=line_first_byte) && (sel_stop_sort>=line_last_byte)) whole_line_selected = true;
        if((!selected) || (sel_stop_sort<line_first_byte) || (sel_start_sort>line_last_byte)) whole_line_not_selected = true;
        if(((aoff+words_per_line-1) < word_count) && (whole_line_selected || whole_line_not_selected)) {
            // draw whole line - hex
            s = "";
            for(i=0;i<words_per_line;i++) {
                if(i>0) s += " ";
                for(j=0;j<bytes_per_word;j++) {
                    buf.Printf("%02X", memory[(aoff+i)*bytes_per_word + j]);
                    s += buf;
                }
            }
            if(whole_line_selected) {
                dcTemp->SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_HIGHLIGHTTEXT));
                dcTemp->SetTextBackground(wxSystemSettings::GetColour(wxSYS_COLOUR_HIGHLIGHT));
                dcTemp->SetBackgroundMode(wxSOLID);
            }
            else {
                dcTemp->SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT));
                dcTemp->SetBackgroundMode(wxTRANSPARENT);
            }
            dcTemp->DrawText(s, character_width*(address_digits+2), character_height*y);
            // draw whole line - text
            s = "";
            for(i=0;i<bytes_per_line;i++) {
                s += GetChar(aoff*bytes_per_word + i);
            }
            dcTemp->DrawText(s, character_width*(address_digits+1+hexlen+3), character_height*y);
        }
        else {
            // draw byte by byte
            int byte_index;
            for(i=0;i<words_per_line;i++) {
                if((aoff+i) >= word_count) break;
                for(j=0;j<bytes_per_word;j++) {
                    byte_index = (aoff+i)*bytes_per_word + j;
                    buf.Printf("%02X", memory[byte_index]);
                    if(selected && (sel_start_sort<=byte_index) && (sel_stop_sort>=byte_index)) {
                        dcTemp->SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_HIGHLIGHTTEXT));
                        dcTemp->SetTextBackground(wxSystemSettings::GetColour(wxSYS_COLOUR_HIGHLIGHT));
                        dcTemp->SetBackgroundMode(wxSOLID);
                        // selected - spaces
                        if((i<(words_per_line-1)) && (j==(bytes_per_word-1)) && (sel_stop_sort>=byte_index+1)) buf += " ";
                    }
                    else {
                        dcTemp->SetTextForeground(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOWTEXT));
                        dcTemp->SetBackgroundMode(wxTRANSPARENT);
                    }
                    dcTemp->DrawText(buf, character_width*(address_digits+2+i*(bytes_per_word*2+1)+j*2), character_height*y);
                    dcTemp->DrawText(GetChar(byte_index), character_width*(address_digits+1+hexlen+3+i*bytes_per_word+j), character_height*y);
                }
            }
        }
        //
        y++;
        aoff += words_per_line;
        if(y>=visible_lines) break;
    }

    // cursor
    int cursor_line = (cursor_pos/2)/bytes_per_line - address_offset/words_per_line;
    if((cursor_line>=0) && (cursor_line<(visible_lines-1))) {
        int cursor_col_n = cursor_pos % (bytes_per_line*2);
        int cursor_col_w = cursor_col_n / (bytes_per_word*2);
        int cursor_col_z = cursor_col_n % (bytes_per_word*2);

	    dcTemp->SetPen(*wxBLACK_PEN);
	    dcTemp->SetBrush(*wxTRANSPARENT_BRUSH);
	    dcTemp->DrawRectangle(character_width*(address_digits+1+cursor_col_w*(bytes_per_word*2+1)+1+(cursor_col_z/2)*2), character_height*(cursor_line+1), character_width*2+1, character_height);
        dcTemp->DrawRectangle(character_width*(address_digits+1+hexlen+3+cursor_col_n/2), character_height*(cursor_line+1), character_width+1, character_height);
        //dcTemp->SetPen(*wxRED_PEN);
        //dcTemp->DrawRectangle(character_width*(address_digits+1+cursor_col_w*(bytes_per_word*2+1)+1+cursor_col_z), character_height*(cursor_line+1), character_width+1, character_height);
    }
	return dcTemp;
}

void HexEdit::MoveCaret()
{
    wxCaret *caret = GetCaret();
    if(!caret) return;
    if(!this->HasFocus()) {
        if(caret->IsVisible()) caret->Hide();
    }
    else {
        int wordpos = (cursor_pos/2) / bytes_per_word;
        if((wordpos >= address_offset) && (wordpos < (address_offset+words_per_line*(visible_lines-1)))) {
            int cursor_col_n = cursor_pos % (bytes_per_line*2);
            int cursor_col_w = cursor_col_n / (bytes_per_word*2);
            int cursor_col_z = cursor_col_n % (bytes_per_word*2);
            int cursor_line = (cursor_pos/2)/bytes_per_line - address_offset/words_per_line;
            int hexlen = (bytes_per_word*2+1)*words_per_line;
            if(cursor_in_hex) {
                caret->Move(character_width*(address_digits+1+cursor_col_w*(bytes_per_word*2+1)+1+cursor_col_z),
                            character_height*(cursor_line+1));
            }
            else {
                caret->Move(character_width*(address_digits+1+hexlen+3+cursor_col_n/2),
                            character_height*(cursor_line+1));
            }
            if(!caret->IsVisible()) caret->Show();
        }
        else {
            if(caret->IsVisible()) caret->Hide();
        }
    }
}

wxChar HexEdit::GetChar(int pos)
{
    wxString s;
    if((memory[pos]>=0x20) && (memory[pos]<=0x7E)) s.Printf("%c", memory[pos]);
    else s = ".";
    return s[0];
}

void HexEdit::OnChar(wxKeyEvent &event)
{
    bool SkipRepaint = false;

	switch (event.GetKeyCode()){
		
        case WXK_LEFT:case WXK_NUMPAD_LEFT:case WXK_RIGHT:case WXK_NUMPAD_RIGHT:
        case WXK_UP:case WXK_NUMPAD_UP:case WXK_DOWN:case WXK_NUMPAD_DOWN:
        case WXK_HOME:case WXK_NUMPAD_HOME:case WXK_END:case WXK_NUMPAD_END:
        case WXK_PAGEUP:case WXK_NUMPAD_PAGEUP:
		case WXK_PAGEDOWN:case WXK_NUMPAD_PAGEDOWN:
            switch (event.GetKeyCode()){
                case WXK_LEFT:case WXK_NUMPAD_LEFT:	        Left();	            break;
		        case WXK_RIGHT:case WXK_NUMPAD_RIGHT:		Right();			break;
		        case WXK_UP:case WXK_NUMPAD_UP:				Up();			    break;
		        case WXK_DOWN:case WXK_NUMPAD_DOWN:			Down(); 			break;
                case WXK_PAGEUP:case WXK_NUMPAD_PAGEUP:     PageUp();           break;
		        case WXK_PAGEDOWN:case WXK_NUMPAD_PAGEDOWN:	PageDown();         break;
		        case WXK_HOME:case WXK_NUMPAD_HOME:
                    if(event.ControlDown()) FileHome();
                    else LineHome();
                    break;
		        case WXK_END:case WXK_NUMPAD_END:
                    if(event.ControlDown()) FileEnd();
                    else LineEnd();
                    break;
            }
            if(event.ShiftDown()) {
                sel_stop = cursor_pos;
                if(sel_start == sel_stop) selected = false;
                else selected = true;
            }
            else {
                sel_start = cursor_pos;
                selected = false;
            }
            break;
		
		case WXK_DELETE:case WXK_NUMPAD_DELETE:
            if(cursor_in_hex) Write("0");
            else Write("\\x00");
            if(!selected) sel_start = cursor_pos;
            MemoryChanged();
			break;
		case WXK_BACK:
            if(cursor_in_hex) Write("0", true);
            else Write("\\x00", true);
            if(!selected) sel_start = cursor_pos;
            MemoryChanged();
			break;

		default:
			wxChar chr = event.GetKeyCode();
            if((chr>=0x20) && (chr<=0x7E) && !event.AltDown() && !event.ControlDown()) {
                Write(wxString(chr));
                if(!selected) sel_start = cursor_pos;
                MemoryChanged();
			}
            // TAB
            else if((chr=='\t') && !event.AltDown() && !event.ControlDown()) {
                cursor_in_hex = !cursor_in_hex;
			}
            // ctrl+c
            else if(!event.AltDown() && event.ControlDown() && !event.ShiftDown() && (chr==WXK_CONTROL_C) ){
                ClipboardCopy();
                SkipRepaint=true;
			}
            // ctrl+v
            else if(!event.AltDown() && event.ControlDown() && !event.ShiftDown() && (chr==WXK_CONTROL_V) ){
                if (wxTheClipboard->Open()){
                    if (wxTheClipboard->IsSupported( wxDF_TEXT )) {
                        wxTextDataObject data;
                        wxTheClipboard->GetData(data);
                        ClipboardPaste(data.GetText());
                        if(!selected) sel_start = cursor_pos;
                    }
                    else {
                        SkipRepaint=true;
                    }
                    wxTheClipboard->Close();
                }
                else {
                    SkipRepaint=true;
                }
			}
            // ctrl+a
            else if(!event.AltDown() && event.ControlDown() && !event.ShiftDown() && (chr==WXK_CONTROL_A) ){
				sel_start = 0;
                FileEnd();
                sel_stop = cursor_pos;
                selected = true;
			}
            else {
                SkipRepaint=true;
            }
			break;
	} //switch end
    if(!SkipRepaint) Repaint();
}

void HexEdit::OnMouseLeft(wxMouseEvent& event)
{
    PointToCoord(event.m_x, event.m_y, &cursor_in_hex, &cursor_pos);
    AdjustAddressOffset();
    if(event.ShiftDown()) {
        sel_stop = cursor_pos;
        if(sel_start == sel_stop) selected = false;
        else selected = true;
    }
    else {
        sel_start = cursor_pos;
        selected = false;
    }
    Repaint();
}

void HexEdit::OnMouseMove(wxMouseEvent& event)
{
    if(event.m_leftDown){
        PointToCoord(event.m_x, event.m_y, &cursor_in_hex, &cursor_pos);
        AdjustAddressOffset();
        sel_stop = cursor_pos;
        if(sel_start == sel_stop) selected = false;
        else selected = true;
        Repaint();
    }
}

void HexEdit::PointToCoord(int x, int y, bool* in_hex, int* position)
{
    int line = (y / character_height) -1 ;
    if(line<0) line = 0;
    if(line>(visible_lines-2)) line=visible_lines-2;
    int line_offset = address_offset + line*words_per_line; // words
    int hexlen = (bytes_per_word*2+1)*words_per_line;
    int middle = character_width*(address_digits+1+hexlen+2);
    if(x<middle) {
        *in_hex = true;
        int xpos = x / character_width;
        xpos -= address_digits+2;
        if(xpos<0) xpos=0;
        if(xpos>=hexlen) xpos=hexlen-1;
        int xword = xpos / (bytes_per_word*2+1);
        int xmod = xpos % (bytes_per_word*2+1);
        if(xmod>=(bytes_per_word*2)) xmod=(bytes_per_word*2)-1;
        int nibblepos = (line_offset+xword)*bytes_per_word*2 + xmod;
        if(nibblepos>=(bytes_count*2)) nibblepos=(bytes_count*2)-1;
        *position = nibblepos;
    }
    else {
        *in_hex = false;
        int xpos = ((x - middle) / character_width) -1;
        if(xpos<0) xpos=0;
        if(xpos>=bytes_per_line) xpos=bytes_per_line-1;
        int bytepos = line_offset*bytes_per_word + xpos;
        if(bytepos>=bytes_count) bytepos=bytes_count-1;
        *position = bytepos*2;
    }
}

void HexEdit::OnFocus(wxFocusEvent& event)
{
    Repaint();
}

void HexEdit::AdjustAddressOffset()
{
    int tmp_word;
    tmp_word = cursor_pos / (bytes_per_word*2);
    if(tmp_word < address_offset) {
        address_offset = (tmp_word/words_per_line)*words_per_line;
    }
    if(tmp_word > (address_offset -1 + (visible_lines-1)*words_per_line)) {
        address_offset = (tmp_word/words_per_line)*words_per_line  - (visible_lines-2)*words_per_line;
    }
    Scroll(0, address_offset/words_per_line);
}

void HexEdit::Left()
{
    if(cursor_in_hex) cursor_pos = cursor_pos - 1;
    else cursor_pos = (cursor_pos/2)*2 - 2;
    if(cursor_pos < 0) cursor_pos = 0;
    AdjustAddressOffset();
}

void HexEdit::Right()
{
    if(cursor_in_hex) cursor_pos = cursor_pos + 1;
    else cursor_pos = (cursor_pos/2)*2 + 2;
    if(cursor_pos > (bytes_count*2 - 1)) cursor_pos = bytes_count*2 - 1;
    AdjustAddressOffset();
}

void HexEdit::Up()
{
    if(cursor_in_hex) cursor_pos = cursor_pos - bytes_per_line*2;
    else cursor_pos = (cursor_pos/2)*2 - bytes_per_line*2;
    if(cursor_pos < 0) cursor_pos = 0;
    AdjustAddressOffset();
}

void HexEdit::Down()
{
    cursor_pos = cursor_pos + bytes_per_line*2;
    if(cursor_pos > (bytes_count*2 - 1)) cursor_pos = (bytes_count*2 - 1);
    if(!cursor_in_hex) cursor_pos = (cursor_pos/2)*2;
    AdjustAddressOffset();
}

void HexEdit::LineHome()
{
    int round = bytes_per_line*2;
    cursor_pos = (cursor_pos/round)*round;
    AdjustAddressOffset();
}

void HexEdit::LineEnd()
{
    int round = bytes_per_line*2;
    cursor_pos = (cursor_pos/round)*round + round - 1;
    if(cursor_pos > (bytes_count*2 - 1)) cursor_pos = (bytes_count*2 - 1);
    AdjustAddressOffset();
}

void HexEdit::FileHome()
{
    cursor_pos = 0;
    AdjustAddressOffset();
}

void HexEdit::FileEnd()
{
    cursor_pos = bytes_count*2 - 1;
    AdjustAddressOffset();
}

void HexEdit::PageUp()
{
    int dispnibbles = bytes_per_line*2 * (visible_lines-1);
    cursor_pos -= dispnibbles;
    if(cursor_pos < 0) cursor_pos = 0;
    address_offset -= words_per_line * (visible_lines-1);
    if(address_offset < 0) address_offset = 0;
    AdjustAddressOffset();
}

void HexEdit::PageDown()
{
    int dispnibbles = bytes_per_line*2 * (visible_lines-1);
    cursor_pos += dispnibbles;
    if(cursor_pos > (bytes_count*2 - 1)) cursor_pos = (bytes_count*2 - 1);
    address_offset += words_per_line * (visible_lines-1);
    int lastlineoffset = address_offset + (visible_lines-2)*words_per_line;
    if(lastlineoffset > (word_count-1)) {
        address_offset = ((word_count-1)/words_per_line)*words_per_line  - (visible_lines-2)*words_per_line;
    }
    AdjustAddressOffset();
}

void HexEdit::ClipboardCopy()
{
    if(!selected) return;
    int sel_start_sort = sel_start;
    int sel_stop_sort = sel_stop;
    if(sel_start > sel_stop) {
        sel_start_sort = sel_stop;
        sel_stop_sort = sel_start;
    }
    sel_start_sort /= 2; // nibbles->bytes
    sel_stop_sort /= 2;
    wxString copytext;
    if(cursor_in_hex) {
        int i, j;
        wxString buf;
        j = 1;
        for(i=sel_start_sort;i<=sel_stop_sort;i++) {
            buf.Printf("%02X", memory[i]);
            copytext += buf;
            if(i<sel_stop_sort) {
                if((j % bytes_per_line) == 0) {
                    copytext += '\n';
                }
                else if((j % bytes_per_word) == 0) {
                    copytext += ' ';
                }
            }
            j++;
        }
    }
    else {
        int i;
        wxString buf;
        for(i=sel_start_sort;i<=sel_stop_sort;i++) {
            if((memory[i]>=0x20) && (memory[i]<=0x7E)) buf.Printf("%c", memory[i]);
            else buf.Printf("\\x%02X", memory[i]);
            copytext += buf;
        }
    }
    // fill into clipboard
    if (wxTheClipboard->Open()) {
        wxTheClipboard->SetData(new wxTextDataObject(copytext));
        wxTheClipboard->Close();
        // This data objects are held by the clipboard,
        // so do not delete them in the app.
    }
}

void HexEdit::ClipboardPaste(wxString text)
{
    wxString pastedata = text;
    if(cursor_in_hex) {
        pastedata.Replace("0x", ""); // lower case
        pastedata.Replace("0X", ""); // upper case
    }
    Write(pastedata);
    MemoryChanged();
}

void HexEdit::Write(wxString s, bool move_left)
{
    if(cursor_in_hex) { // write into HEX
        int i;
        wxString s_hex;
        for(i=0;i<(int)s.Len();i++) {
            if(ToHex(s[i]) >=0) s_hex += s[i];
        }
        if(s_hex.IsEmpty()) return;
        if(selected) {
            int sel_start_sort = sel_start;
            int sel_stop_sort = sel_stop;
            if(sel_start > sel_stop) {
                sel_start_sort = sel_stop;
                sel_stop_sort = sel_start;
            }
            sel_start_sort /= 2; // nibbles->bytes
            sel_stop_sort /= 2;
            wxString s_ext = s_hex;
            while((int)s_ext.Len() < ((sel_stop_sort-sel_start_sort+1)*2)) s_ext += s_hex;
            int k = 0;
            for(i=sel_start_sort;i<=sel_stop_sort;i++) {
                memory[i] = (ToHex(s_ext[k++]) << 4) | ToHex(s_ext[k++]);
            }
        }
        else {
            for(i=0;i<(int)s_hex.Len();i++) {
                if((cursor_pos % 2) == 0) {
                    memory[cursor_pos/2] = (memory[cursor_pos/2] & 0x0F) | (ToHex(s_hex[i]) << 4);
                }
                else {
                    memory[cursor_pos/2] = (memory[cursor_pos/2] & 0xF0) | ToHex(s_hex[i]);
                }
                if(move_left) {
                    if(cursor_pos==0) break;
                    Left();
                }
                else {
                    if(cursor_pos==(bytes_count*2-1)) break;
                    Right();
                }
            }
        }
    }
    else { // write into TEXT
        int i;
        // convert string to bytes
        wxMemoryBuffer mbuf;
        i=0;
        while(i<(int)s.Len()) {
            if((s[i]=='\\') && (i<((int)s.Len()-3)) && (s[i+1]=='x') && (ToHex(s[i+2])>=0) && (ToHex(s[i+3])>=0)) {
                mbuf.AppendByte((ToHex(s[i+2])<<4) | ToHex(s[i+3]));
                i += 4;
            }
            else if(s[i]<=0xFF) {
                mbuf.AppendByte(s[i]);
                i += 1;
            }
            else {
                mbuf.AppendByte((char)0xFF);
                i += 1;
            }
        }
        if(mbuf.GetDataLen()==0) return;
        if(selected) {
            int sel_start_sort = sel_start;
            int sel_stop_sort = sel_stop;
            if(sel_start > sel_stop) {
                sel_start_sort = sel_stop;
                sel_stop_sort = sel_start;
            }
            sel_start_sort /= 2; // nibbles->bytes
            sel_stop_sort /= 2;
            wxMemoryBuffer xmbuf;
            while((int)xmbuf.GetDataLen() < (sel_stop_sort-sel_start_sort+1)) {
                xmbuf.AppendData(mbuf.GetData(), mbuf.GetDataLen());
            }
            char* pxmbuf = (char*)xmbuf.GetData();
            int k = 0;
            for(i=sel_start_sort;i<=sel_stop_sort;i++) {
                memory[i] = pxmbuf[k++];
            }
        }
        else {
            char* pmbuf = (char*)mbuf.GetData();
            for(i=0;i<(int)mbuf.GetDataLen();i++) {
                memory[cursor_pos/2] = pmbuf[i];
                if(move_left) {
                    if(cursor_pos<=1) break;
                    Left();
                }
                else {
                    if(cursor_pos>=(bytes_count*2-2)) break;
                    Right();
                }
            }
        }
    }
}

int HexEdit::ToHex(wxUniChar hex)
{
	return ( hex >= '0' && hex <= '9' ) ? hex -'0' :
			( hex >= 'a' && hex <= 'f' ) ? hex -'a' + 10:
			( hex >= 'A' && hex <= 'F' ) ? hex -'A' + 10:
			-1;
}

void HexEdit::SetCallback(wxObject* dst, EditorUpdateHandler handler)
{
    update_dst = dst;
    update_callback = handler;
    update_callback_set = true;
}

void HexEdit::MemoryChanged(void)
{
    if(update_callback_set) {
        (*update_dst.*update_callback)();
    }
}


